@@ -4,6 +4,7 @@
 from __future__ import print_function
 from ._compat import PY2, range_type, text_type, str_type, JYTHON, IRONPYTHON
 
+import errno
 from functools import partial
 import os
 import hashlib
@@ -231,52 +232,26 @@ def unpack_int64(n):
 
 
 class MysqlPacket(object):
-    """Representation of a MySQL response packet.  Reads in the packet
-    from the network socket, removes packet header and provides an interface
-    for reading/parsing the packet results."""
+    """Representation of a MySQL response packet.
 
-    def __init__(self, connection):
+    Provides an interface for reading/parsing the packet results.
+    """
+    __slots__ = ('_position', '_data')
+
+    def __init__(self, data, encoding):
         self._position = 0
-        self._recv_packet(connection)
-
-    def _recv_packet(self, connection):
-        """Parse the packet header and read entire packet payload into buffer."""
-        buff = b''
-        while True:
-            packet_header = connection._read_bytes(4)
-            if len(packet_header) < 4:
-                raise OperationalError(2013, "Lost connection to MySQL server during query")
-
-            if DEBUG: dump_packet(packet_header)
-            packet_length_bin = packet_header[:3]
-
-            #TODO: check sequence id
-            self.__packet_number = byte2int(packet_header[3])
-
-            bin_length = packet_length_bin + b'\0'  # pad little-endian number
-            bytes_to_read = struct.unpack('<I', bin_length)[0]
-            recv_data = connection._read_bytes(bytes_to_read)
-            if len(recv_data) < bytes_to_read:
-                raise OperationalError(2013, "Lost connection to MySQL server during query")
-            if DEBUG: dump_packet(recv_data)
-            buff += recv_data
-            if bytes_to_read < MAX_PACKET_LEN:
-                break
-        self.__data = buff
-
-    def packet_number(self):
-        return self.__packet_number
+        self._data = data
 
     def get_all_data(self):
-        return self.__data
+        return self._data
 
     def read(self, size):
         """Read the first 'size' bytes in packet and advance cursor past them."""
-        result = self.__data[self._position:(self._position+size)]
+        result = self._data[self._position:(self._position+size)]
         if len(result) != size:
             error = ('Result length not requested length:\n'
                      'Expected=%s.  Actual=%s.  Position: %s.  Data Length: %s'
-                     % (size, len(result), self._position, len(self.__data)))
+                     % (size, len(result), self._position, len(self._data)))
             if DEBUG:
                 print(error)
                 self.dump()
@@ -289,21 +264,21 @@ class MysqlPacket(object):
 
         (Subsequent read() will return errors.)
         """
-        result = self.__data[self._position:]
+        result = self._data[self._position:]
         self._position = None  # ensure no subsequent read()
         return result
 
     def advance(self, length):
         """Advance the cursor in data buffer 'length' bytes."""
         new_position = self._position + length
-        if new_position < 0 or new_position > len(self.__data):
+        if new_position < 0 or new_position > len(self._data):
             raise Exception('Invalid advance amount (%s) for cursor.  '
                             'Position=%s' % (length, new_position))
         self._position = new_position
 
     def rewind(self, position=0):
         """Set the position of the data buffer cursor to 'position'."""
-        if position < 0 or position > len(self.__data):
+        if position < 0 or position > len(self._data):
             raise Exception("Invalid position to rewind cursor to: %s." % position)
         self._position = position
 
@@ -316,7 +291,7 @@ class MysqlPacket(object):
         No error checking is done.  If requesting outside end of buffer
         an empty string (or string shorter than 'length') may be returned!
         """
-        return self.__data[position:(position+length)]
+        return self._data[position:(position+length)]
 
     def read_length_encoded_integer(self):
         """Read a 'Length Coded Binary' number from the data buffer.
@@ -349,31 +324,31 @@ class MysqlPacket(object):
         return self.read(length)
 
     def is_ok_packet(self):
-        return self.__data[0:1] == b'\0'
+        return self._data[0:1] == b'\0'
 
     def is_eof_packet(self):
         # http://dev.mysql.com/doc/internals/en/generic-response-packets.html#packet-EOF_Packet
         # Caution: \xFE may be LengthEncodedInteger.
         # If \xFE is LengthEncodedInteger header, 8bytes followed.
-        return len(self.__data) < 9 and self.__data[0:1] == b'\xfe'
+        return len(self._data) < 9 and self._data[0:1] == b'\xfe'
 
     def is_resultset_packet(self):
-        field_count = ord(self.__data[0:1])
+        field_count = ord(self._data[0:1])
         return 1 <= field_count <= 250
 
     def is_error_packet(self):
-        return self.__data[0:1] == b'\xff'
+        return self._data[0:1] == b'\xff'
 
     def check_error(self):
         if self.is_error_packet():
             self.rewind()
             self.advance(1)  # field_count == error (we already know that)
             errno = unpack_uint16(self.read(2))
-            if DEBUG: print("errno = {}".format(errno))
-            raise_mysql_exception(self.__data)
+            if DEBUG: print("errno =", errno)
+            raise_mysql_exception(self._data)
 
     def dump(self):
-        dump_packet(self.__data)
+        dump_packet(self._data)
 
 
 class FieldDescriptorPacket(MysqlPacket):
@@ -383,9 +358,9 @@ class FieldDescriptorPacket(MysqlPacket):
     attributes on the class such as: db, table_name, name, length, type_code.
     """
 
-    def __init__(self, connection):
-        MysqlPacket.__init__(self, connection)
-        self.__parse_field_descriptor(connection.encoding)
+    def __init__(self, data, encoding):
+        MysqlPacket.__init__(self, data, encoding)
+        self.__parse_field_descriptor(encoding)
 
     def __parse_field_descriptor(self, encoding):
         """Parse the 'Field Descriptor' (Metadata) packet.
@@ -405,27 +380,19 @@ class FieldDescriptorPacket(MysqlPacket):
         self.flags = struct.unpack('<H', self.read(2))[0]
         self.scale = byte2int(self.read(1))  # "decimals"
         self.advance(2)  # filler (always 0x00)
-
         # 'default' is a length coded binary and is still in the buffer?
         # not used for normal result sets...
 
     def description(self):
         """Provides a 7-item tuple compatible with the Python PEP249 DB Spec."""
-        desc = []
-        desc.append(self.name)
-        desc.append(self.type_code)
-        desc.append(None) # TODO: display_length; should this be self.length?
-        desc.append(self.get_column_length()) # 'internal_size'
-        desc.append(self.get_column_length()) # 'precision'  # TODO: why!?!?
-        desc.append(self.scale)
-
-        # 'null_ok' -- can this be True/False rather than 1/0?
-        #              if so just do:  desc.append(bool(self.flags % 2 == 0))
-        if self.flags % 2 == 0:
-            desc.append(1)
-        else:
-            desc.append(0)
-        return tuple(desc)
+        return (
+            self.name,
+            self.type_code,
+            None,  # TODO: display_length; should this be self.length?
+            self.get_column_length(),  # 'internal_size'
+            self.get_column_length(),  # 'precision'  # TODO: why!?!?
+            self.scale,
+            self.flags % 2 == 0)
 
     def get_column_length(self):
         if self.type_code == FIELD_TYPE.VAR_STRING:
@@ -459,12 +426,10 @@ class OKPacketWrapper(object):
         self.server_status = struct.unpack('<H', self.packet.read(2))[0]
         self.warning_count = struct.unpack('<H', self.packet.read(2))[0]
         self.message = self.packet.read_all()
+        self.has_next = self.server_status & SERVER_STATUS.SERVER_MORE_RESULTS_EXISTS
 
     def __getattr__(self, key):
-        if hasattr(self.packet, key):
-            return getattr(self.packet, key)
-        raise AttributeError(
-            "{0} instance has no attribute '{1}'".format(self.__class__, key))
+        return getattr(self.packet, key)
 
 
 class EOFPacketWrapper(object):
@@ -488,10 +453,7 @@ class EOFPacketWrapper(object):
         self.has_next = self.server_status & SERVER_STATUS.SERVER_MORE_RESULTS_EXISTS
 
     def __getattr__(self, key):
-        if hasattr(self.packet, key):
-            return getattr(self.packet, key)
-        raise AttributeError(
-            "{0} instance has no attribute '{1}'".format(self.__class__, key))
+        return getattr(self.packet, key)
 
 
 class Connection(object):
@@ -505,21 +467,21 @@ class Connection(object):
 
     socket = None
 
-    def __init__(self, host="localhost", user=None, passwd="",
+    def __init__(self, host="localhost", user=None, password="",
                  database=None, port=3306, unix_socket=None,
                  charset='', sql_mode=None,
                  read_default_file=None, conv=decoders, use_unicode=None,
                  client_flag=0, cursorclass=Cursor, init_command=None,
                  connect_timeout=None, ssl=None, read_default_group=None,
                  compress=None, named_pipe=None, no_delay=False,
-                 autocommit=False, db=None):
+                 autocommit=False, db=None, passwd=None):
         """
         Establish a connection to the MySQL database. Accepts several
         arguments:
 
         host: Host where the database server is located
         user: Username to log in as
-        passwd: Password to use.
+        password: Password to use.
         database: Database to use, None to not use a particular one.
         port: MySQL port to use, default is usually OK.
         unix_socket: Optionally, you can use a unix socket rather than TCP/IP.
@@ -538,7 +500,9 @@ class Connection(object):
         named_pipe: Not supported
         no_delay: Disable Nagle's algorithm on the socket
         autocommit: Autocommit mode. None means use server default. (default: False)
+
         db: Alias for database. (for compatibility to MySQLdb)
+        passwd: Alias for password. (for compatibility to MySQLdb)
         """
 
         if use_unicode is None and sys.version_info[0] > 2:
@@ -546,6 +510,8 @@ class Connection(object):
 
         if db is not None and database is None:
             database = db
+        if passwd is not None and not password:
+            password = passwd
 
         if compress or named_pipe:
             raise NotImplementedError("compress and named_pipe arguments are not supported")
@@ -585,7 +551,7 @@ class Connection(object):
                     return default
 
             user = _config("user", user)
-            passwd = _config("password", passwd)
+            password = _config("password", password)
             host = _config("host", host)
             database = _config("database", database)
             unix_socket = _config("socket", unix_socket)
@@ -595,7 +561,7 @@ class Connection(object):
         self.host = host
         self.port = port
         self.user = user or DEFAULT_USER
-        self.password = passwd
+        self.password = password or ""
         self.db = database
         self.no_delay = no_delay
         self.unix_socket = unix_socket
@@ -640,7 +606,7 @@ class Connection(object):
         send_data = struct.pack('<i', 1) + int2byte(COM_QUIT)
         try:
             self._write_bytes(send_data)
-        except IOError:
+        except Exception:
             pass
         finally:
             sock = self.socket
@@ -648,9 +614,18 @@ class Connection(object):
             self._rfile = None
             sock.close()
 
+    @property
+    def open(self):
+        return self.socket is not None
+
     def __del__(self):
         if self.socket:
-            self.close()
+            try:
+                self.socket.close()
+            except:
+                pass
+        self.socket = None
+        self._rfile = None
 
     def autocommit(self, value):
         self.autocommit_mode = bool(value)
@@ -665,7 +640,7 @@ class Connection(object):
     def _read_ok_packet(self):
         pkt = self._read_packet()
         if not pkt.is_ok_packet():
-            raise OperationalErrorl(2014, "Command Out of Sync")
+            raise OperationalError(2014, "Command Out of Sync")
         ok = OKPacketWrapper(pkt)
         self.server_status = ok.server_status
         return True
@@ -676,6 +651,11 @@ class Connection(object):
                               self.escape(self.autocommit_mode))
         self._read_ok_packet()
 
+    def begin(self):
+        """Begin transaction."""
+        self._execute_command(COM_QUERY, "BEGIN")
+        self._read_ok_packet()
+
     def commit(self):
         ''' Commit changes to stable storage '''
         self._execute_command(COM_QUERY, "COMMIT")
@@ -734,8 +714,8 @@ class Connection(object):
         self._affected_rows = self._read_query_result(unbuffered=unbuffered)
         return self._affected_rows
 
-    def next_result(self):
-        self._affected_rows = self._read_query_result()
+    def next_result(self, unbuffered=False):
+        self._affected_rows = self._read_query_result(unbuffered=unbuffered)
         return self._affected_rows
 
     def affected_rows(self):
@@ -778,16 +758,23 @@ class Connection(object):
         try:
             if self.unix_socket and self.host in ('localhost', '127.0.0.1'):
                 sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-                t = sock.gettimeout()
                 sock.settimeout(self.connect_timeout)
                 sock.connect(self.unix_socket)
-                sock.settimeout(t)
                 self.host_info = "Localhost via UNIX socket"
                 if DEBUG: print('connected using unix_socket')
             else:
-                sock = socket.create_connection((self.host, self.port), self.connect_timeout)
+                while True:
+                    try:
+                        sock = socket.create_connection(
+                                (self.host, self.port), self.connect_timeout)
+                        break
+                    except (OSError, IOError) as e:
+                        if e.errno == errno.EINTR:
+                            continue
+                        raise
                 self.host_info = "socket %s:%d" % (self.host, self.port)
                 if DEBUG: print('connected using socket')
+            sock.settimeout(None)
             sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
             if self.no_delay:
                 sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
@@ -820,17 +807,49 @@ class Connection(object):
 
     def _read_packet(self, packet_type=MysqlPacket):
         """Read an entire "mysql packet" in its entirety from the network
-        and return a MysqlPacket type that represents the results."""
+        and return a MysqlPacket type that represents the results.
+        """
+        buff = b''
+        while True:
+            packet_header = self._read_bytes(4)
+            if DEBUG: dump_packet(packet_header)
+            packet_length_bin = packet_header[:3]
 
-        packet = packet_type(self)
+            #TODO: check sequence id
+            #  packet_number
+            byte2int(packet_header[3])
+
+            bin_length = packet_length_bin + b'\0'  # pad little-endian number
+            bytes_to_read = struct.unpack('<I', bin_length)[0]
+            recv_data = self._read_bytes(bytes_to_read)
+            if DEBUG: dump_packet(recv_data)
+            buff += recv_data
+            if bytes_to_read < MAX_PACKET_LEN:
+                break
+        packet = packet_type(buff, self.encoding)
         packet.check_error()
         return packet
 
     def _read_bytes(self, num_bytes):
-        return self._rfile.read(num_bytes)
+        while True:
+            try:
+                data = self._rfile.read(num_bytes)
+                break
+            except (IOError, OSError) as e:
+                if e.errno == errno.EINTR:
+                    continue
+                raise OperationalError(
+                        2013, "Lost connection to MySQL server during query (%r)" % (e,))
+        if len(data) < num_bytes:
+            raise OperationalError(2013,
+                    "Lost connection to MySQL server during query")
+        return data
 
     def _write_bytes(self, data):
-        self.socket.sendall(data)
+        try:
+            self.socket.sendall(data)
+        except IOError as e:
+            raise OperationalError(2006, "MySQL server has gone away (%r)" % (e,))
 
     def _read_query_result(self, unbuffered=False):
         if unbuffered:
@@ -839,6 +858,7 @@ class Connection(object):
                 result.init_unbuffered_query()
             except:
                 result.unbuffered_active = False
+                result.connection = None
                 raise
         else:
             result = MySQLResult(self)
@@ -917,9 +937,10 @@ class Connection(object):
                                           ssl_version=ssl.PROTOCOL_TLSv1,
                                           cert_reqs=ssl.CERT_REQUIRED,
                                           ca_certs=self.ca)
+            self._rfile = _makefile(self.socket, 'rb')
 
         data = data_init + self.user + b'\0' + \
-            _scramble(self.password.encode(self.encoding), self.salt)
+            _scramble(self.password.encode('latin1'), self.salt)
 
         if self.db:
             if isinstance(self.db, text_type):
@@ -933,24 +954,17 @@ class Connection(object):
 
         self._write_bytes(data)
 
-        auth_packet = MysqlPacket(self)
-        auth_packet.check_error()
-        if DEBUG: auth_packet.dump()
+        auth_packet = self._read_packet()
 
         # if old_passwords is enabled the packet will be 1 byte long and
         # have the octet 254
 
         if auth_packet.is_eof_packet():
             # send legacy handshake
-            #raise NotImplementedError, "old_passwords are not supported. Check to see if mysqld was started with --old-passwords, if old-passwords=1 in a my.cnf file, or if there are some short hashes in your mysql.user table."
-            # TODO: is this the correct charset?
-            data = _scramble_323(self.password.encode(self.encoding), self.salt.encode(self.encoding)) + b'\0'
+            data = _scramble_323(self.password.encode('latin1'), self.salt) + b'\0'
             data = pack_int24(len(data)) + int2byte(next_packet) + data
-
             self._write_bytes(data)
-            auth_packet = MysqlPacket(self)
-            auth_packet.check_error()
-            if DEBUG: auth_packet.dump()
+            auth_packet = self._read_packet()
 
     # _mysql support
     def thread_id(self):
@@ -967,8 +981,7 @@ class Connection(object):
 
     def _get_server_information(self):
         i = 0
-        packet = MysqlPacket(self)
-        packet.check_error()
+        packet = self._read_packet()
         data = packet.get_all_data()
 
         if DEBUG: dump_packet(data)
@@ -1028,8 +1041,7 @@ class Connection(object):
 class MySQLResult(object):
 
     def __init__(self, connection):
-        from weakref import proxy
-        self.connection = proxy(connection)
+        self.connection = connection
         self.affected_rows = None
         self.insert_id = None
         self.server_status = None
@@ -1046,13 +1058,16 @@ class MySQLResult(object):
             self._finish_unbuffered_query()
 
     def read(self):
-        first_packet = self.connection._read_packet()
+        try:
+            first_packet = self.connection._read_packet()
 
-        # TODO: use classes for different packet types?
-        if first_packet.is_ok_packet():
-            self._read_ok_packet(first_packet)
-        else:
-            self._read_result_packet(first_packet)
+            # TODO: use classes for different packet types?
+            if first_packet.is_ok_packet():
+                self._read_ok_packet(first_packet)
+            else:
+                self._read_result_packet(first_packet)
+        finally:
+            self.connection = False
 
     def init_unbuffered_query(self):
         self.unbuffered_active = True
@@ -1061,6 +1076,7 @@ class MySQLResult(object):
         if first_packet.is_ok_packet():
             self._read_ok_packet(first_packet)
             self.unbuffered_active = False
+            self.connection = None
         else:
             self.field_count = first_packet.read_length_encoded_integer()
             self._get_descriptions()
@@ -1077,6 +1093,7 @@ class MySQLResult(object):
         self.server_status = ok_packet.server_status
         self.warning_count = ok_packet.warning_count
         self.message = ok_packet.message
+        self.has_next = ok_packet.has_next
 
     def _check_packet_is_eof(self, packet):
         if packet.is_eof_packet():
@@ -1100,6 +1117,7 @@ class MySQLResult(object):
         packet = self.connection._read_packet()
         if self._check_packet_is_eof(packet):
             self.unbuffered_active = False
+            self.connection = None
             self.rows = None
             return
 
@@ -1116,6 +1134,7 @@ class MySQLResult(object):
             packet = self.connection._read_packet()
             if self._check_packet_is_eof(packet):
                 self.unbuffered_active = False
+                self.connection = None  # release reference to kill cyclic reference.
 
     def _read_rowdata_packet(self):
         """Read a rowdata packet for each data row in the result set."""
@@ -1123,6 +1142,7 @@ class MySQLResult(object):
         while True:
             packet = self.connection._read_packet()
             if self._check_packet_is_eof(packet):
+                self.connection = None  # release reference to kill cyclic reference.
                 break
             rows.append(self._read_row_from_packet(packet))
 
