@@ -1,9 +1,15 @@
+import pymysql.cursors
+
 from pymysql.tests import base
 from pymysql import util
+from pymysql.err import ProgrammingError
 
 import time
 import datetime
 
+__all__ = ["TestConversion", "TestCursor", "TestBulkInserts"]
+
+
 class TestConversion(base.PyMySQLTestCase):
     def test_datatypes(self):
         """ test every data type """
@@ -102,16 +108,36 @@ class TestConversion(base.PyMySQLTestCase):
         c.execute("select '',null")
         self.assertEqual((u'',None), c.fetchone())
 
-    def test_datetime(self):
-        """ test conversion of null, empty string """
+    def test_timedelta(self):
+        """ test timedelta conversion """
         conn = self.connections[0]
         c = conn.cursor()
-        c.execute("select time('12:30'), time('23:12:59'), time('23:12:59.05100')")
+        c.execute("select time('12:30'), time('23:12:59'), time('23:12:59.05100'), time('-12:30'), time('-23:12:59'), time('-23:12:59.05100'), time('-00:30')")
         self.assertEqual((datetime.timedelta(0, 45000),
                           datetime.timedelta(0, 83579),
-                          datetime.timedelta(0, 83579, 51000)),
+                          datetime.timedelta(0, 83579, 51000),
+                          -datetime.timedelta(0, 45000),
+                          -datetime.timedelta(0, 83579),
+                          -datetime.timedelta(0, 83579, 51000),
+                          -datetime.timedelta(0, 1800)),
                          c.fetchone())
 
+    def test_datetime(self):
+        """ test datetime conversion """
+        conn = self.connections[0]
+        c = conn.cursor()
+        dt = datetime.datetime(2013,11,12,9,9,9,123450)
+        try:
+            c.execute("create table test_datetime (id int, ts datetime(6))")
+            c.execute("insert into test_datetime values (1,'2013-11-12 09:09:09.12345')")
+            c.execute("select ts from test_datetime")
+            self.assertEqual((dt,),c.fetchone())
+        except ProgrammingError:
+            # User is running a version of MySQL that doesn't support msecs within datetime
+            pass
+        finally:
+            c.execute("drop table if exists test_datetime")
+
 
 class TestCursor(base.PyMySQLTestCase):
     # this test case does not work quite right yet, however,
@@ -207,7 +233,82 @@ class TestCursor(base.PyMySQLTestCase):
             c.execute("drop table mystuff")
 
 
-__all__ = ["TestConversion","TestCursor"]
+class TestBulkInserts(base.PyMySQLTestCase):
+
+    cursor_type = pymysql.cursors.DictCursor
+
+    def setUp(self):
+        super(TestBulkInserts, self).setUp()
+        self.conn = conn = self.connections[0]
+        c = conn.cursor(self.cursor_type)
+
+        # create a table ane some data to query
+        c.execute("drop table if exists bulkinsert")
+        c.execute(
+"""CREATE TABLE bulkinsert
+(
+id int(11),
+name char(20),
+age int,
+height int,
+PRIMARY KEY (id)
+)
+""")
+
+    def _verify_records(self, data):
+        conn = self.connections[0]
+        cursor = conn.cursor()
+        cursor.execute("SELECT id, name, age, height from bulkinsert")
+        result = cursor.fetchall()
+        self.assertEqual(sorted(data), sorted(result))
+
+    def test_bulk_insert(self):
+        conn = self.connections[0]
+        cursor = conn.cursor()
+
+        data = [(0, "bob", 21, 123), (1, "jim", 56, 45), (2, "fred", 100, 180)]
+        cursor.executemany("insert into bulkinsert (id, name, age, height) "
+                           "values (%s,%s,%s,%s)", data)
+        self.assertEqual(
+            cursor._last_executed, bytearray(
+            b"insert into bulkinsert (id, name, age, height) values "
+            b"(0,'bob',21,123),(1,'jim',56,45),(2,'fred',100,180)"))
+        cursor.execute('commit')
+        self._verify_records(data)
+
+    def test_bulk_insert_multiline_statement(self):
+        conn = self.connections[0]
+        cursor = conn.cursor()
+        data = [(0, "bob", 21, 123), (1, "jim", 56, 45), (2, "fred", 100, 180)]
+        cursor.executemany("""insert
+into bulkinsert (id, name,
+age, height)
+values (%s,
+%s , %s,
+%s )
+ """, data)
+        self.assertEqual(cursor._last_executed, bytearray(b"""insert
+into bulkinsert (id, name,
+age, height)
+values (0,
+'bob' , 21,
+123 ),(1,
+'jim' , 56,
+45 ),(2,
+'fred' , 100,
+180 )"""))
+        cursor.execute('commit')
+        self._verify_records(data)
+
+    def test_bulk_insert_single_record(self):
+        conn = self.connections[0]
+        cursor = conn.cursor()
+        data = [(0, "bob", 21, 123)]
+        cursor.executemany("insert into bulkinsert (id, name, age, height) "
+                           "values (%s,%s,%s,%s)", data)
+        cursor.execute('commit')
+        self._verify_records(data)
+
 
 if __name__ == "__main__":
     import unittest
